#include "alarmmonitor.h"
#include "alarmbusitem.h"

#include <velib/qt/charger_error.hpp>
#include <velib/qt/lynxion_error.hpp>
#include <velib/qt/vebus_error.hpp>

AlarmMonitor::AlarmMonitor(DBusService *service, Type type,
			const QString &busitemPathAlarm, const QString &alarmEnablePath,
			const QString &alarmValuePath, DeviceAlarms *parent) :
	QObject(parent), mService(service), mNotification(0), mAlarmTrigger(0), mType(type), mDeviceAlarms(parent)
{
	// Alarms can optionally be enabled / supressed by a setting.
	if (!alarmEnablePath.isEmpty()) {
		mEnabledNotifications = NO_ALARM;
		// FIXME: get this from the settings provider
		QString id("dbus/com.victronenergy.settings/Settings/Alarm" + alarmEnablePath);
		VeQItems::getRoot()->itemGetOrCreate(id)->getValueAndChanges(this, SLOT(settingChanged(VeQItem *, QVariant)));
	} else {
		mEnabledNotifications = ALARM_AND_WARNING;
	}

	// Optionally an value can be associated with an alarm, e.g. voltage
	// for an low voltage alarm.
	if (!alarmValuePath.isEmpty()) {
		mBusitemValue = service->item(alarmValuePath);
		mBusitemValue->getText();
	} else {
		mBusitemValue = new VeQItem(0, this);
	}

	// The actual trigger of the alarm, see Type for the supported formats.
	mAlarmTrigger = service->item(busitemPathAlarm);
	mAlarmTrigger->getValueAndChanges(this, SLOT(updateAlarm(VeQItem *, QVariant)));
}

AlarmMonitor::~AlarmMonitor()
{
	if (mNotification)
		mNotification->setActive(false);
}

bool AlarmMonitor::mustBeShown(DbusAlarm alarm)
{
	if (alarm == DBUS_NO_ERROR)
		return false;

	if (mEnabledNotifications == ALARM_ONLY && alarm == DBUS_WARNING)
		return false;

	return true;
}

void AlarmMonitor::updateAlarm(VeQItem *item, QVariant var)
{
	Q_UNUSED(item);

	// If there was a previous warning / error it is no longer valid
	if (mNotification) {
		mNotification->setActive(false);
		mNotification->disconnect(this);
		mNotification = 0;
	}

	if (!var.isValid() || mEnabledNotifications == NO_ALARM)
		return;

	DbusAlarm alarm = DBUS_NO_ERROR;
	switch (mType)
	{
	case REGULAR:
		alarm = static_cast<DbusAlarm>(var.toInt());
		break;
	case VEBUS_ERROR:
	{
		int vebusErrorCode = var.toInt();
		if (vebusErrorCode == 0) {
			alarm = DBUS_NO_ERROR;
		} else {
			alarm = DBUS_ERROR;
			mDescription = VebusError::getDescription(vebusErrorCode);
		}
		break;
	}
	case LYNXION_ERROR:
	{
		int lynxIonErrorCode = var.toInt();
		if (lynxIonErrorCode == 0) {
			alarm = DBUS_NO_ERROR;
		} else {
			alarm = DBUS_ERROR;
			mDescription = LynxIonError::getDescription(lynxIonErrorCode);
		}
		break;
	}
	case CHARGER_ERROR:
	{
		int error = var.toInt();
		if (error == 0) {
			alarm = DBUS_NO_ERROR;
		} else {
			alarm = ChargerError::isWarning(error) ? DBUS_WARNING : DBUS_ERROR;
			mDescription = ChargerError::getDescription(error);
		}
		break;
	}
	default:
		break;
	}

	if (!mustBeShown(alarm))
		return;

	if (alarm == DBUS_WARNING)
		addNotification(Notification::WARNING);
	else
		addNotification(Notification::ALARM);
}

// Copied to mEnabledNotifications since having the enabled as a setting is optional
void AlarmMonitor::settingChanged(VeQItem *item, QVariant var)
{
	mEnabledNotifications = static_cast<Enabled>(var.toInt());
	if (mAlarmTrigger)
		updateAlarm(item, mAlarmTrigger->getValue());
}

void AlarmMonitor::addNotification(Notification::Type type)
{
	mNotification = mDeviceAlarms->notificationCenter()->addNotification(type, mService->getDescription(),
														  mDescription, mBusitemValue->getText());
	connect(mNotification, SIGNAL(destroyed()), this, SLOT(notificationDestroyed()));
}

void AlarmMonitor::notificationDestroyed()
{
	mNotification = 0;
}
