#include <iostream>

#include <application.hpp>

using namespace QsLogging;
using namespace std;

/*
 * qt4 embedded needs the GuiServer flag for QWS to start. QWS support is removed from qt5
 * and a different constructor is called, hence the different types. (Type versus bool)
 */
#if QT_VERSION < QT_VERSION_CHECK(5, 0, 0)
QApplication::Type appType = QApplication::GuiServer;
#else
bool appType = true;
#endif

Application::Application(int &argc, char **argv) :
	QApplication(argc, argv, appType),
	mPlatform(VePlatform::create(*this)),
	mCompileFlags(0)
{
	setCompileFlags();
	setApplicationVersion(VERSION);
	initArguments();

	// Setup logging early, so debug output makes it to the console
	initLogger(QsLogging::TraceLevel);
	if (mArguments.contains("d"))
		Logger::instance().setLoggingLevel(static_cast<QsLogging::Level>(mArguments.value("d").toInt()));

	mPlatform->init(*this, mArguments);

	run();
}

void Application::setCompileFlags()
{
#if defined(CFG_WITH_QWACS)
	mCompileFlags |= CompiledFlagQwacs;
#endif
#if !defined(CFG_WITHOUT_SCREENSAVER)
	mCompileFlags |= CompiledFlagScreenSaver;
#endif
}

void Application::initLogger(QsLogging::Level logLevel)
{
	// init the logging mechanism
	DestinationPtr debugDestination(DestinationFactory::MakeDebugOutputDestination() );
	Logger *logger = &Logger::instance();
	logger->addDestination(debugDestination);
	logger->setIncludeTimestamp(false);

	cout << "GUI " << "v" VERSION << " started" << endl;
	cout << "Built with Qt " << QT_VERSION_STR << " running on " << qVersion() << endl;
	logger->setLoggingLevel(logLevel);
}

void Application::initArguments()
{
	mArguments.addArg("-h", "Print this help");
	mArguments.addArg("-d level", "Debug level: 0=TRACE, 1=DEBUG, 2=INFO...");
}

void Application::run()
{
	if (mArguments.contains("h")) {
		mArguments.help();
		exit(0);
	}

	mDefaultView = new VeQuickView(mPlatform, 0 /* &mItemRoot */);
	connect(mPlatform, SIGNAL(upAndRunning()), SLOT(onUpAndRunning()), Qt::QueuedConnection);
	mPlatform->run(mArguments);
}

void Application::onUpAndRunning()
{
	disconnect(mPlatform, SIGNAL(upAndRunning()));

	loadTranslation();
	mDefaultView->qmlRegister();
	mDefaultView->loadApplication();

#if 0
	/* open a second CCGX view */
	VeQuickView *test = new VeQuickView(mPlatform, 0 /* &mItemRoot */);
	test->qmlRegister();
	test->loadApplication();
	test->show();
#endif
}

void Application::loadTranslation()
{
	QString language = mPlatform->getLanguage();

	if (language == "en")
		return;

	QString qmFile(qApp->applicationDirPath() + "/lang/color_control_" + language + ".qm");
	if (mTranslator.load(qmFile)) {
		qApp->installTranslator(&mTranslator);
	} else {
		QLOG_ERROR() << "[" << __func__ << __FILE__ << ":" << __LINE__ << "] Failed to load translation file: " << qmFile;
	}
}
