#include <QFile>
#include <QTextStream>
#include <QDebug>
#ifdef Q_WS_QWS
#include <QWSServer>
#endif
#include <QDateTime>
#include <QMetaType>
#include <QtNetwork/QNetworkInterface>

#include "colorcontrol.h"
#include "QsLog.h"
#include "connman/cmagent.h"

#include <velib/qt/charger_error.hpp>
#include <velib/qt/v_busitems.h>

ColorControl::ColorControl(QDeclarativeView *view, VeQItemSettings *settings, QObject *parent) :
	QObject(parent),
	mView(view),
	mLanguage("en"),
	mDemoMode(false),
	mBrightness(100),
	mBooting(true),
	mCmdExecuting(false),
	mAccessLevel(AccessUser),
	mLocalSettings(new VSettings(VBusItems::getConnection(), this)),
	mBuzzer("dbus/com.victronenergy.system/Buzzer/State"),
	mAudibleAlarmObtained(false),
#ifdef CFG_WITH_QWACS
	mQwacs(view->rootContext()),
#endif
	mConnman(0)
{
	getVersionInfo();
	QLOG_INFO() << "CCGX version" << mVersion << "build on " << mVersionDate;
	QLOG_INFO() << "GUI built on" << mBuildDate << "at" << mBuildTime;

	qmlRegisterUncreatableType<ColorControl>("com.victron.velib", 1, 0, "Victron", ""); // to access types
	mView->rootContext()->setContextProperty("ColorControl", this);

	mCmdline = new QProcess(this);
	connect(mCmdline, SIGNAL(finished(int, QProcess::ExitStatus)), this, SLOT(cmdlineFinished(int, QProcess::ExitStatus)));

	// Brigthness
	mItemBrightness = settings->root()->itemGetOrCreate("Settings/Gui/Brightness");
	mItemBrightness->getValueAndChanges(this, SLOT(brightnessSettingChanged(VeQItem*,QVariant)));

	// Screensaver
	mDbusDisplayOffTime = settings->root()->itemGetOrCreate("Settings/Gui/DisplayOff");
	mDbusDisplayOffTime->getValueAndChanges(this, SLOT(displayOffTimeSettingChanged(VeQItem*,QVariant)));

	mDBusServices = new DBusServices(this);
	mView->rootContext()->setContextProperty("DBusServices", mDBusServices);

	// Daemontools daemons
	new Console(settings, "/service/vegetty", "Settings/Services/Console", parent);
	new Service(settings, "/service/dbus-modbustcp", "Settings/Services/Modbus", parent);
	new Service(settings, "/service/mosquitto", "Settings/Services/Mqtt", parent);
	new Service(settings, "/service/dbus-mqtt", "Settings/Services/Mqtt", parent);
	Service *motordrive = new Service(settings, "/service/dbus-motordrive", "Settings/Services/OceanvoltMotorDrive", parent);
	Service *valence = new Service(settings, "/service/dbus-valence", "Settings/Services/OceanvoltValence", parent);
	Service *resu = new Service(settings, "/service/lg-resu-interface", "Settings/Services/LgResu", parent);
	Service *vecan = new Service(settings, "/service/vecan-dbus", "Settings/Services/VeCan", parent);
	new Service(settings, "/service/sshd", "Settings/System/RemoteSupport", parent);
	new Service(settings, "/service/dbus-generator-starter", "Settings/Relay/Function", parent);
	new Service(settings, "/service/dbus-pump-starter", "Settings/Relay/Function", 3, parent);

	QList<QString> list = QList<QString>() << "Settings/System/RemoteSupport" << "Settings/System/VncInternet";
	new Service(settings, "/service/ssh-tunnel", list, parent);

	new Service(settings, "/service/vrmpubnub", "Settings/Services/Vrmpubnub", parent);
	new Service(settings, "/service/websockify-c", "Settings/System/VncLocal", parent);

	// The resu has a bitrate of 500k, the others 250k, so mark the as conflicting (and hence the are turned off)
	resu->addConflictingService(motordrive);
	resu->addConflictingService(valence);
	resu->addConflictingService(vecan);

	// Alarms
	mNotificationCenter = new NotificationCenter(this);
	mView->rootContext()->setContextProperty("NotificationCenter", mNotificationCenter);
	connect(mNotificationCenter, SIGNAL(alarmChanged()), this, SLOT(alarmChanged()));
	mAlarmBusitem = new AlarmBusitem(mDBusServices, mNotificationCenter);

	mAudibleAlarm = settings->root()->itemGetOrCreate("Settings/Alarm/Audible");
	mAudibleAlarm->getValueAndChanges(this, SLOT(audibleAlarmSettingChanged(VeQItem*,QVariant)));

	// Updater
	mUpdater = new Updater(mNotificationCenter, this);
	mView->rootContext()->setContextProperty("Updater", mUpdater);

	// Relay
	mRelay = new Relay("dbus/com.victronenergy.system/Relay/0/State", mNotificationCenter, this);
	mView->rootContext()->setContextProperty("Relay", mRelay);

	// Connman
	mConnman = CmManager::instance(this);
	mView->rootContext()->setContextProperty("Connman", mConnman);
	qmlRegisterType<CmTechnology>("net.connman",0,1,"CmTechnology");
	qmlRegisterType<CmService>("net.connman",0,1,"CmService");
	qmlRegisterType<CmAgent>("net.connman",0,1,"CmAgent");
	qmlRegisterType<ClockModel>("net.connman",0,1,"ClockModel");

	qmlRegisterType<ChargerError>("com.victron.velib", 1, 0, "ChargerError");

	// TimeZone
	mTimeZone = new TimeZone(settings, this);
	mView->rootContext()->setContextProperty("TimeZone", mTimeZone);

	// Access level
	mDbusAccessLevel = settings->root()->itemGetOrCreate("Settings/System/AccessLevel");
	mDbusAccessLevel->getValueAndChanges(this, SLOT(accessLevelSettingChanged(VeQItem*,QVariant)));

	// Demo mode
	mDemoModeSetting = settings->root()->itemGetOrCreate("Settings/Gui/DemoMode");
	mDemoModeSetting->getValueAndChanges(this, SLOT(demoSettingChanged(VeQItem*,QVariant)));

	// Umount signal form dbus
	VBusItems::getConnection().connect("","/","com.victronenergy.udev" ,"umount",
											   this, SLOT(diskUnmounted(const QString &, const QString &, const ushort)));

	// VRM logger storagelocation dbus signal
	VBusItems::getConnection().connect("","","com.victronenergy.logger" ,"storagelocation",
											   this, SLOT(vrmStorageLocation(const uint, const qulonglong)));

	// VRM logger storageerror dbus signal
	VBusItems::getConnection().connect("","","com.victronenergy.logger" ,"storageerror",
											   this, SLOT(vrmStorageError(const uint)));

	// With all slots hooked up, scan for services
	mDBusServices->initialScan();
	mBooting = false;
}

void ColorControl::getVersionInfo()
{
	QFile file("/opt/color-control/version");

	mVersion = "---";
	mVersionDate =  "---";
	mBuildDate = __DATE__;
	mBuildTime = __TIME__;
	if (!file.open(QIODevice::ReadOnly | QIODevice::Text))
		return;

	mVersion = file.readLine().trimmed();
	mVersionDate = file.readLine(); // Dummy read
	mVersionDate = file.readLine().trimmed();
	file.close();
}

QString ColorControl::getVersion()
{
	return mVersion;
}

QString ColorControl::getVersionDate()
{
	return mVersionDate;
}

ColorControl::AccessLevel ColorControl::getAccessLevel()
{
	return mAccessLevel;
}

void ColorControl::setAccessLevel(AccessLevel level)
{
	if (mAccessLevel == level)
		return;

	QLOG_TRACE() << "ColorControl::setAccessLevel: " << level;
	mAccessLevel = level;
	emit accessLevelChanged();
}

void ColorControl::setBrightness(int level)
{
	if (level < 15) {
		QLOG_WARN() << "cowardly refusing to set the backlight below 15, since it is not visible";
		level = 15;
	}

	if (level == mBrightness)
		return;

	mBrightness = level;

	QFile file("/sys/class/backlight/bpp3-bl/brightness");
	if (!file.open(QIODevice::WriteOnly | QIODevice::Text))
		return;

	QTextStream out(&file);
	out << QString::number(level);
	QLOG_TRACE() << "ColorControl::setBacklight" << mBrightness;
	file.close();

	emit brightnessChanged();
}

int ColorControl::getBrightness()
{
	return mBrightness;
}

void ColorControl::brightnessSettingChanged(VeQItem *item, QVariant var)
{
	Q_UNUSED(item);

	if (var.isValid())
		setBrightness(var.toInt());
}

QString ColorControl::getLanguage()
{
	return mLanguage;
}

void ColorControl::setLanguage(QString &lang)
{
	mLanguage = lang;
}

void ColorControl::changeLanguage(QString lang)
{
	if (mLanguage != lang) {
		splash(tr("Changing language..."));
		qApp->quit();
	}
}

bool ColorControl::wakeupEvent()
{
	if (mScreenSaver.isSaved()) {
		mScreenSaver.enable();
		return true;
	}

	mScreenSaver.trigger();
	return false;
}

bool ColorControl::silenceBuzzer()
{
	if (mBuzzer.isBeeping()) {
		mBuzzer.buzzerOff();
		return true;
	}

	return false;
}

void ColorControl::displayOffTimeSettingChanged(VeQItem *item, QVariant var)
{
	Q_UNUSED(item);

	if (var.isValid()) {
		unsigned int displayOffTime = var.toUInt();
		mScreenSaver.setInhibitTime(displayOffTime);
		QLOG_TRACE() << "ColorControl::dbusDisplayOffTimeChanged: " << displayOffTime;
	}
}

void ColorControl::accessLevelSettingChanged(VeQItem *item, QVariant var)
{
	Q_UNUSED(item);

	if (var.isValid()) {
		bool ok;
		int level = var.toInt(&ok);

		if (!ok || !ColorControl::validAccessLevel(level)) {
			QLOG_ERROR() << "ColorControl::dbusAccessLevelChanged: invalid access level" << level;
			return;
		}

		QLOG_TRACE() << "ColorControl::dbusAccessLevelChanged: " << level;
		setAccessLevel(static_cast<AccessLevel>(level));
	}
}

void ColorControl::alarmChanged()
{
	if (mNotificationCenter->getAlarm()) {
		if (mAudibleAlarm->getValue().isValid() && mAudibleAlarm->getValue().toBool())
			mBuzzer.buzzerOn();
		else
			mBuzzer.buzzerOff();
		mScreenSaver.disable();
	} else {
		mScreenSaver.enable();
		mBuzzer.buzzerOff();
	}
}

void ColorControl::audibleAlarmSettingChanged(VeQItem *item, QVariant var)
{
	Q_UNUSED(item);

	if (var.isValid()) {
		if (!mAudibleAlarmObtained)
			alarmChanged();
		mAudibleAlarmObtained = true;
	}
}

void ColorControl::diskUnmounted(const QString &devicename, const QString &mountpoint, const ushort result)
{
	Q_UNUSED(devicename);
	Q_UNUSED(mountpoint);
	if (result == 1) {
		emit uncleanDiskUnmount();
		wakeupEvent();
	}
}

void ColorControl::vrmStorageLocation(const uint location, const qulonglong freeBytes)
{
	// External storage
	if (location == 2) {
		emit vrmStorageAttached(freeBytes);
		wakeupEvent();
	}
}

void ColorControl::vrmExternalStorageError(const uint error)
{
	switch(error)
	{
	case 1:
	case 4:
		emit vrmStorageError(error);
		wakeupEvent();
		break;
	}
}

QString ColorControl::getUTCDateTime()
{
	QDateTime dateTime;
	dateTime = QDateTime::currentDateTimeUtc();
	return dateTime.toString("yyyy-MM-dd hh:mm");
}

QString ColorControl::getLocalDateTime()
{
	QDateTime dateTime;
	dateTime = QDateTime::currentDateTime();
	return dateTime.toString("yyyy-MM-dd hh:mm");
}

QString ColorControl::getMacAddress()
{
	QNetworkInterface interface = QNetworkInterface::interfaceFromName("eth0");
	return interface.hardwareAddress();
}

QString ColorControl::getVRMPortalID()
{
	QString macAddress = getMacAddress();
	QString portalID = macAddress.toLower();
	portalID.replace(":", "");

	return portalID;
}

// Dummy
bool ColorControl::getDoReboot()
{
	return true;
}

void ColorControl::setDoReboot(bool reboot)
{
	QLOG_TRACE() << "ColorControl::setDoReboot(" << reboot << ")";
	splash(tr("Rebooting..."));
	mCmdline->start("/opt/color-control/gui/reboot.sh");
	mCmdline->waitForFinished(300000);
}

bool ColorControl::getDemoMode()
{
	return mDemoMode;
}

void ColorControl::setDemoMode(bool demoMode)
{
	if (demoMode != mDemoMode) {
		mDemoMode = demoMode;
		setCmdExecuting(true);
		if (mDemoMode) {
			QLOG_TRACE() << "ColorControl::setDemoMode start";
			mCmdline->start("/opt/color-control/dbusrecorder/startdemo.sh");
		} else {
			QLOG_TRACE() << "ColorControl::setDemoMode stop";
			QString command = "/opt/color-control/dbusrecorder/stopdemo.sh";
			command += mBooting ? " booting" : "";
			mCmdline->start(command);
		}
	}
}

void ColorControl::demoSettingChanged(VeQItem *item, QVariant var)
{
	Q_UNUSED(item);

	if (var.isValid())
		setDemoMode(var.toBool());
}

bool ColorControl::getCmdExecuting()
{
	return mCmdExecuting;
}

QString ColorControl::setRootPassword(QString password)
{
#ifdef TARGET_ccgx
	QString cmd = "echo \"root:" + password.replace("\"", "\\\"") + "\" | chpasswd";
	int ret = system(cmd.toLatin1());

	if (ret == 0)
		return "Root password changed to " + password;
	else
		return "Changing password failed";
#else
	Q_UNUSED(password)
	return "Can't change root password (not a ccgx)";
#endif
}

void ColorControl::setCmdExecuting(bool isExecuting)
{
	if (isExecuting != mCmdExecuting) {
		mCmdExecuting = isExecuting;
		emit cmdExecutingChanged();
	}
}

void ColorControl::cmdlineFinished(int exitCode, QProcess::ExitStatus exitStatus)
{
	Q_UNUSED(exitCode);
	Q_UNUSED(exitStatus);
	QLOG_TRACE() << "ColorControl::cmdlineFinished";
	setCmdExecuting(false);
}

void ColorControl::splash(const QString &message)
{
	QPixmap exitSplash("/opt/color-control/gui/splash.png");
	if (exitSplash.isNull())
		return;
	QPainter pixPaint(&exitSplash);
	QFont font(pixPaint.font());

	font.setPointSize(32);
	font.setBold(true);
	pixPaint.setFont(font);
	pixPaint.setPen(Qt::white);
	QRect rect=pixPaint.boundingRect(0, 0, 480, 272, Qt::AlignHCenter, message);
	pixPaint.drawText((480 - rect.width()) / 2, 240, message);

#if TARGET_ccgx
	QWSServer::setBackground(exitSplash);
#endif
}

void ColorControl::showStart()
{
	splash(tr("Starting..."));
}
