#ifndef COLORCONTROL_H
#define COLORCONTROL_H

#include <QDeclarativeView>
#include <QProcess>
#include <QKeyEvent>
#include <QPixmap>

#include "QsLog.h"
#include "v_settings.h"
#include "buzzer.h"
#include "dbus_services.h"
#include "notification_center.h"
#include "alarmbusitem.h"
#include "screensaver.h"
#include "updater.h"
#include "relay.h"
#include "timezone.h"
#include <qwacs/qwacs.h>
#include "service.h"
#include "connman/cmmanager.h"
#include <ve_qitems_settings.hpp>

class ColorControl : public QObject
{
	Q_OBJECT
	Q_ENUMS(AccessLevel)
	Q_ENUMS(FirmwareType)
	Q_PROPERTY(QString version READ getVersion NOTIFY dummyNotifiedChanged)
	Q_PROPERTY(QString versionDate READ getVersionDate NOTIFY dummyNotifiedChanged)
	Q_PROPERTY(int brightness READ getBrightness WRITE setBrightness NOTIFY brightnessChanged)
	Q_PROPERTY(QString language READ getLanguage WRITE changeLanguage NOTIFY dummyNotifiedChanged)
	Q_PROPERTY(QString utcDateTime READ getUTCDateTime NOTIFY dummyNotifiedChanged)
	Q_PROPERTY(QString localDateTime READ getLocalDateTime NOTIFY dummyNotifiedChanged)
	Q_PROPERTY(QString macAddress READ getMacAddress NOTIFY dummyNotifiedChanged)
	Q_PROPERTY(QString portalID READ getVRMPortalID NOTIFY dummyNotifiedChanged)
	Q_PROPERTY(bool doReboot READ getDoReboot WRITE setDoReboot)
	Q_PROPERTY(AccessLevel accessLevel READ getAccessLevel WRITE setAccessLevel NOTIFY accessLevelChanged)
	Q_PROPERTY(bool cmdExecuting READ getCmdExecuting NOTIFY cmdExecutingChanged)

public:
	ColorControl(QDeclarativeView *view, VeQItemSettings *settings, QObject *parent = 0);

	enum AccessLevel {
		AccessUser,
		AccessInstaller,
		AccessSuperUser,
		AccessService
	};

	enum FirmwareType {
		FirmwareRelease,
		FirmwareCandidate,
		FirmwareTesting,
		FirmwareDevelop
	};

	static bool validAccessLevel(int val) {
		return (val >= AccessUser && val <= AccessService);
	}

	static void showStart();

	QString getVersion();
	QString getVersionDate();
	void setBrightness(int level);
	int getBrightness();
	QString getLanguage();
	void changeLanguage(QString lang);
	void setLanguage(QString &lang);
	bool getDemoMode();
	void setDemoMode(bool demoMode);
	QString getUTCDateTime();
	QString getLocalDateTime();
	QString getMacAddress();
	QString getVRMPortalID();
	bool getDoReboot();
	void setDoReboot(bool reboot);
	AccessLevel getAccessLevel();
	void setAccessLevel(AccessLevel level);
	bool getCmdExecuting();
	Q_INVOKABLE QString setRootPassword(QString password);

	bool wakeupEvent();
	bool silenceBuzzer();

signals:
	void dummyNotifiedChanged();
	void accessLevelChanged();
	void cmdExecutingChanged();
	void brightnessChanged();
	void uncleanDiskUnmount();
	void vrmStorageAttached(qlonglong availableSpace);
	void vrmStorageError(uint error);

private slots:
	void alarmChanged();
	void displayOffTimeSettingChanged(VeQItem *item, QVariant var);
	void accessLevelSettingChanged(VeQItem *item, QVariant var);
	void brightnessSettingChanged(VeQItem *item, QVariant var);
	void demoSettingChanged(VeQItem *item, QVariant var);
	void audibleAlarmSettingChanged(VeQItem *item, QVariant var);
	void cmdlineFinished(int, QProcess::ExitStatus);
	void diskUnmounted(const QString &devicename, const QString &mountpoint, const ushort result);
	void vrmStorageLocation(const uint location, const qulonglong freeBytes);
	void vrmExternalStorageError(const uint error);

private:
	void getVersionInfo();
	void initLogger(QsLogging::Level logLevel);
	static void splash(const QString &message);
	void writeLanguage(QString &lang);
	void setCmdExecuting(bool isExecuting);

	QDeclarativeView *mView;
	QString mVersion;
	QString mVersionDate;
	QString mBuildDate;
	QString mBuildTime;
	QString mLanguage;
	bool mDemoMode;
	int mBrightness;
	bool mBooting;
	bool mCmdExecuting;
	AccessLevel mAccessLevel;

	QProcess *mCmdline;
	VSettings *mLocalSettings;

	ScreenSaver mScreenSaver;
	Buzzer mBuzzer;
	NotificationCenter *mNotificationCenter;
	AlarmBusitem *mAlarmBusitem;
	Updater *mUpdater;
	Relay *mRelay;
	TimeZone *mTimeZone;
	DBusServices *mDBusServices;

	VeQItem *mDbusDisplayOffTime;
	VeQItem *mDbusAccessLevel;
	VeQItem *mItemBrightness;
	VeQItem *mDemoModeSetting;
	VeQItem *mAudibleAlarm;
	bool mAudibleAlarmObtained;

#ifdef CFG_WITH_QWACS
	Qwacs mQwacs;
#endif

	CmManager *mConnman;

	// Snapshot privates
	QPixmap mOriginalPixmap;
};

Q_DECLARE_METATYPE(ColorControl::AccessLevel);
Q_DECLARE_METATYPE(ColorControl::FirmwareType);

#endif // COLORCONTROL_H
