#include <QCoreApplication>
#include <QsLog.h>
#ifdef Q_WS_QWS
#include <QWSServer>
#endif

extern "C" {
	#include <ow-crypt.h>
}

#include <application.hpp>
#include <platform/ve_platform_ccgx.hpp>
#include <velib/qt/v_busitems.h>
#include <ve_qitem_table_widget.hpp>

#include <colorcontrol.h>

void VePlatformCcgx::init(Application &app, Arguments &args)
{
	Q_UNUSED(app);

	// Default to the system bus on the ccgx since there is no session bus
	// note: the actual connection is checked in run...
	VBusItems::setConnectionType(QDBusConnection::SystemBus);
	args.addArg("--dbus address", "dbus address or 'session' or 'system'");

#ifdef Q_WS_QWS
	app.setProperty("NoMStyle", true);
	QWSServer::setCursorVisible(false);
#endif
}

void VePlatformCcgx::qmlRegister(VeQuickView *view)
{
	VePlatform::qmlRegister(view);

	mColorControl = new ColorControl(view, mSettings);
	QString language = getLanguage();
	mColorControl->setLanguage(language);
	mColorControl->showStart();
}

bool VePlatformCcgx::wakeupEvent()
{
	return mColorControl->wakeupEvent();
}

bool VePlatformCcgx::silenceBuzzer()
{
	return mColorControl->silenceBuzzer();
}

/*
 * note: no attempt is made to clean the password from the heap, stack.
 * It cannot be done reliably in cpp without knowing the implementation
 * of a QString (toUtf8 might make a copy e.g. an attempt to clear it might
 * just set the length / copy on write etc).
 */
#define BCRYPT_HASHSIZE			64

bool VePlatformCcgx::setRemoteConsolePassword(const QString &password)
{
	/*
	 * The salt is only to make the encryted password device specific, it is not a
	 * secret. Min size is 16 so it is added twice (it is far from random, since it
	 * is the MAC address). Since the MAC address is used, it looks hex, but the
	 * ASCII values are used, not the BCD. The "setting" starts with the base64 version
	 * of the salt, so while it contains the salt, it is not a one on one copy.
	 */
	QString portalId(mColorControl->getVRMPortalID() + mColorControl->getVRMPortalID());

	char const *salt = portalId.toLatin1().constData();
	size_t saltLength = strlen(salt);
	char setting[BCRYPT_HASHSIZE]; /* out, contains the salt */
	if (crypt_gensalt_rn("$2a$", 8, salt, saltLength, setting, BCRYPT_HASHSIZE) == NULL) {
		qWarning() << "creating setting from salt failed" << portalId;
		return false;
	}

	// note: toUtf8 is used in an attempt to support special characters
	char const *passwd = password.toUtf8().constData();
	char hash[BCRYPT_HASHSIZE];
	char const *aux = crypt_rn(passwd, setting, hash, BCRYPT_HASHSIZE);
	if (aux == NULL) {
		qWarning() << "hashing the password failed";
		return false;
	}

	// ignore the hash for an empty password -> no password means anyone is welcome
	if (password.isEmpty())
		hash[0] = 0;

	/*
	 * FIXME: the filename is in arguments at the moment, but not in the Arguments
	 * object somehow, hardcoded for now...
	 */
	QFile passwordFile("/data/conf/vncpassword.txt");
	if (!passwordFile.open(QIODevice::WriteOnly | QIODevice::Text)) {
		qWarning() << "storing password hash failed";
		return false;
	}
	QTextStream out(&passwordFile);
	out << hash << endl;

	return true;
}

void VePlatformCcgx::showView(VeQuickView *view)
{
#if 1
	view->showFullScreen();
#else
	view->show();
#endif
}

void VePlatformCcgx::run(Arguments &args)
{
	if (args.contains("dbus"))
		VBusItems::setDBusAddress(args.value("dbus"));

	QDBusConnection dbus = VBusItems::getConnection();
	if (!dbus.isConnected()) {
		QLOG_ERROR() << "DBus connection failed.";
		exit(EXIT_FAILURE);
	}

#if 0
	// widget to show all values available on the ccgx
	VeQItemTableWidget *widget = new VeQItemTableWidget(mApp.getVeQItemRoot());
	widget->show();
#endif

	emit upAndRunning();
}

QString VePlatformCcgx::getLanguage()
{
	QString language = qgetenv("GUI_LANG");
	if (language.isEmpty())
		language = "en";
	return language;
}

QString VePlatformCcgx::themePath()
{
	QDir mThemeDir(mApp.applicationDirPath() + "/../themes/" + "ccgx");
	mThemeDir.makeAbsolute();
	return mThemeDir.exists() ? mThemeDir.path() : "";
}

QString VePlatformCcgx::qmlUrl(QString path)
{
	QString rootPath;

	rootPath = mApp.applicationDirPath();

	return rootPath + QDir::separator() + path;
}
