#include "qwacs.h"
#include "QsLog.h"
#include <velib/qt/v_busitems.h>

const QString service = "com.victronenergy.settings";
const QString objectPath = "/Settings/Sensors/OnPosition/";
const QString ACIn1_L1 = "ACIn1_L1";
const QString ACIn1_L2 = "ACIn1_L2";
const QString ACIn1_L3 = "ACIn1_L3";
const QString ACIn2_L1 = "ACIn2_L1";
const QString ACIn2_L2 = "ACIn2_L2";
const QString ACIn2_L3 = "ACIn2_L3";
const QString ACOut_L1 = "ACOut_L1";
const QString ACOut_L2 = "ACOut_L2";
const QString ACOut_L3 = "ACOut_L3";

Qwacs::Qwacs(QDeclarativeContext *context, QObject *parent) :
	QObject(parent),
	mManager(0),
	mGateway(0),
	mQwacsWatcher("com.victronenergy.qwacs", VBusItems::getConnection(), QDBusServiceWatcher::WatchForOwnerChange)
{
	//mBlink = false;
	//mBlinking = false;
	mUplink = false;

	qwacsRegisterd("com.victronenergy.qwacs");
	connect(&mQwacsWatcher, SIGNAL(serviceRegistered(QString)), this, SLOT(qwacsRegisterd(QString)));
	connect(&mQwacsWatcher, SIGNAL(serviceUnregistered(QString)), this, SLOT(qwacsUnregisterd(QString)));

	addSettingBusItem(ACIn1_L1);
	addSettingBusItem(ACIn1_L2);
	addSettingBusItem(ACIn1_L3);
	addSettingBusItem(ACIn2_L1);
	addSettingBusItem(ACIn2_L2);
	addSettingBusItem(ACIn2_L3);
	addSettingBusItem(ACOut_L1);
	addSettingBusItem(ACOut_L2);
	addSettingBusItem(ACOut_L3);

	mContext = context;
	mContext->setContextProperty("Qwacs", this);
}

void Qwacs::addSettingBusItem(const QString & item)
{
	BusItemCons * busItem = new BusItemCons(service, objectPath+item, VBusItems::getConnection());

	QString replyString = busItem->getValue().toString();
	QStringList sensors = replyString.split(",");
	for (int i = 0; i < sensors.size(); i++ ) {
		mPositionsMap.insert(sensors[i], item);
	}
	mSensorsMap[item] = replyString;
	mBusItemMap.insert(item, busItem);
}

QStringList Qwacs::getSensors()
{
	QLOG_INFO() << "[Qwacs::getSensors] ";
	return mSensorList;
}

void Qwacs::setPosition(const QString & id, const QString &pos)
{
	if ( (id.isEmpty()) || (mPositionsMap[id] == pos) )
		return;

	QLOG_TRACE() << "[Qwacs::setPosition] pos = " << pos << " id = " << id;
	mPositionsMap[id] = pos;

	// Remove sensor from position
	QMap<QString,QString>::iterator i = mSensorsMap.begin();
	while (i != mSensorsMap.end()) {
		if (i.value().contains(id) ) {
			QLOG_INFO() << "[Qwacs::setPosition] remove " << id << " from " << i.value();
			i.value().remove(id);
			if (i.value().endsWith(','))
				i.value().chop(1);
			if (i.value().startsWith(','))
				i.value().remove(0,1);
			mBusItemMap[i.key()]->setValue(i.value());
			QLOG_INFO() << "[Qwacs::setPosition] " << id << " removed. Result " << i.value();
		}
		++i;
	}

	if (!pos.isEmpty()) {
		// Add sensor to position
		QString sensors;
		QMap<QString,QString>::const_iterator j = mPositionsMap.constBegin();
		while (j != mPositionsMap.constEnd()) {
			if ( (j.value() == pos) ) {
				sensors.append(j.key());
				sensors.append(',');
			}
			++j;
		}
		sensors.chop(1);
		mSensorsMap[pos] = sensors;
		mBusItemMap[pos]->setValue(sensors);
	}
	emit positionChanged();
}

void Qwacs::blinkOn(const QString & id)
{
	mGateway->blinkOn(id);
}

void Qwacs::blinkOff(const QString & id)
{
	mGateway->blinkOff(id);
}

void Qwacs::qwacsRegisterd(const QString serviceName)
{
	QLOG_TRACE() << "[Qwacs::qwacsRegisterd] " << serviceName;
	if (!mManager) {
		mManager = new QwacsManagerInterface(serviceName, "/Manager", VBusItems::getConnection(), this);
		if (mManager->isValid()) {
			connect(mManager, SIGNAL(PropertiesChanged(const QVariantMap &)), this, SLOT(managerChanges(QVariantMap)));
			connect(mManager, SIGNAL(gatewayFound(QString)), this, SLOT(gatewayFound(QString)));
			connect(mManager, SIGNAL(sensorAdded(QString)), this, SLOT(sensorAdded(QString)));
			mVersion = mManager->getProperty("Version").toString();
			mConnected = mManager->getProperty("GatewayConnected").toBool();
			if (mConnected) {
				gatewayFound(mManager->getProperty("Hostname").toString());
				mSensorList = mManager->getProperty("SensorList").toStringList();
				mSensorList.prepend("Gateway");
				emit sensorsChanged();
			}
		} else {
			delete mManager;
			mManager = 0;
			QLOG_TRACE() << "[Qwacs::qwacsRegisterd] service not registerd " << serviceName;
		}
	}
}

void Qwacs::qwacsUnregisterd(const QString serviceName)
{
	QLOG_TRACE() << "[Qwacs::qwacsUnregisterd] " << serviceName;
	if (mManager) {
		disconnect(mManager, SIGNAL(PropertiesChanged(const QVariantMap &)), this, SLOT(managerChanges(QVariantMap)));
		disconnect(mManager, SIGNAL(gatewayFound(QString)), this, SLOT(gatewayFound(QString)));
		disconnect(mManager, SIGNAL(sensorAdded(QString)), this, SLOT(sensorAdded(QString)));
		mVersion = "";
		mConnected = false;
		mSensorList.clear();
		emit sensorsChanged();
		delete mManager;
		mManager = 0;
	}
	if (mGateway) {
		disconnect(mGateway, SIGNAL(PropertiesChanged(const QVariantMap &)), this, SLOT(gatewayChanges(QVariantMap &)));
		delete mGateway;
		mGateway = 0;
	}
}

void Qwacs::gatewayFound(const QString &hostname)
{
	QLOG_TRACE() << "[Qwacs::gatewayFound] " << hostname;
	mHostname = hostname;
	if (!mGateway) {
		mGateway = new QwacsGatewayInterface("com.victronenergy.qwacs", "/Gateway", VBusItems::getConnection(), this);
		connect(mGateway, SIGNAL(PropertiesChanged(const QVariantMap &)), this, SLOT(gatewayChanges(const QVariantMap &)));
		connect(mGateway, SIGNAL(UplinkStatus(const QString &)), this, SLOT(uplinkStatus(const QString &)));
		mCommonName = mGateway->getProperty("commonName").toString();
		mFirmwareVersion = mGateway->getProperty("firmwareVersion").toString();
		mArchFlavVers = mGateway->getProperty("archFlavVers").toString();
		mSerialNr = mGateway->getProperty("serialNr").toString();
		mPartNr = mGateway->getProperty("partNr").toString();
		mUpdays = mGateway->getProperty("updays").toUInt();
		mUphours = mGateway->getProperty("uphours").toUInt();
		if (!mSensorList.contains("Gateway"))
			mSensorList.prepend("Gateway");
	}
}

void Qwacs::sensorAdded(const QString &id)
{
	QLOG_TRACE() << "[Qwacs::sensorAdded] " << id;
	if (!mSensorList.contains(id)) {
		mSensorList.append(id);
		emit sensorsChanged();
	}
}

void Qwacs::managerChanges(const QVariantMap &changes)
{
	Q_UNUSED(changes);
	QLOG_TRACE() << "[Qwacs::managerChanges] ";
}

void Qwacs::gatewayChanges(const QVariantMap &changes)
{
	QLOG_TRACE() << "[Qwacs::gatewayChanges] ";
	QMapIterator<QString, QVariant> i(changes);

	while (i.hasNext()) {
		i.next();
		QLOG_TRACE() << "Item: " << i.key() << ":" << i.value().toString();
		if (i.value().isValid()) {
			if (i.key() == "commonName") {
				if (i.value() != mCommonName) {
					mCommonName = i.value().toString();
					emit commonNameChanged();
				}
			} else if (i.key() == "firmwareVersion") {
				if (i.value() != mFirmwareVersion) {
					mFirmwareVersion = i.value().toString();
					emit firmwareVersionChanged();
				}
			} else if (i.key() == "archFlavVers") {
				if (i.value() != mArchFlavVers) {
					mArchFlavVers = i.value().toString();
					emit archFlavVersChanged();
				}
			} else if (i.key() == "serialNr") {
				if (i.value() != mSerialNr) {
					mSerialNr = i.value().toString();
					emit serialNrChanged();
				}
			} else if (i.key() == "partNr") {
				if (i.value() != mPartNr) {
					mPartNr = i.value().toString();
					emit partNrChanged();
				}
			} else if (i.key() == "updays") {
				if (i.value() != mUpdays) {
					mUpdays = i.value().toUInt();
					emit updaysChanged();
				}
			} else if (i.key() == "uphours") {
				if (i.value() != mUphours) {
					mUphours = i.value().toUInt();
					emit uphoursChanged();
				}
			}
		}
	}
}

void Qwacs::uplinkStatus(const QString & status)
{
	if (status == mUplinkStatus)
		return;

	mUplinkStatus = status;
	mUplink = status == "disabled" ? false : true;
	emit uplinkStatusChanged();
}
