#include <QFile>
#include <QProcess>

#include "service.h"
#include "QsLog.h"

#include <ve_qitems_settings.hpp>

Service::Service(VeQItemSettings *settings, const QString &servicePath, const QList<QString> &dbusObjectPathList, QObject *parent) :
	QObject(parent),
	mServicePath(servicePath),
	mDbusObjectPathList(dbusObjectPathList),
	mStartValue(1)
{
	init(settings);
}

Service::Service(VeQItemSettings *settings, const QString &servicePath, const QString &dbusObjectPath, QObject *parent) :
	QObject(parent),
	mServicePath(servicePath),
	mStartValue(1)
{
	mDbusObjectPathList.append(dbusObjectPath);

	init(settings);
}

Service::Service(VeQItemSettings *settings, const QString &servicePath, const QString &dbusObjectPath, int startValue, QObject *parent):
	QObject(parent),
	mServicePath(servicePath),
	mStartValue(startValue)
{
	mDbusObjectPathList.append(dbusObjectPath);

	init(settings);
}

void Service::init(VeQItemSettings *settings)
{
	QList<QString>::iterator i;
	for (i = mDbusObjectPathList.begin(); i != mDbusObjectPathList.end(); ++i) {
		VeQItem *v = settings->root()->itemGetOrCreate(*i);
		mDbusItemList.append(v);
		v->getValueAndChanges(this, SLOT(dbusItemChanged(VeQItem*,QVariant)));
	}
}

void Service::enableDaemon()
{
	if (daemonIsRunning() && mDbusObjectPathList.count() > 1) {
		// Multiple objects paths control this service. One of them was added or removed: restart the service.
		restartDaemon();
		return;
	}

	if (daemonIsRunning())
		return;

	QLOG_INFO() << "[Service] Start" << mServicePath;

	beforeDaemonStarted();

	QFile::remove(mServicePath + "/down");
	QFile::remove(mServicePath + "/log/down");

	execSvc(QStringList() << "-u" << mServicePath + "/log" << mServicePath);
}

void Service::disableDaemon()
{
	if (!daemonIsRunning())
		return;

	QLOG_INFO() << "[Service] Stop" << mServicePath;

	touch(mServicePath + "/down");
	touch(mServicePath + "/log/down");

	execSvc(QStringList() << "-d" << mServicePath << mServicePath + "/log");
	afterDaemonStopped();
}

void Service::restartDaemon()
{
	QLOG_INFO() << "[Service] Restart" << mServicePath;

	execSvc(QStringList() << "-t" << mServicePath);
}

void Service::addConflictingService(Service *service)
{
	QLOG_TRACE() << "[Service] marking" << service->mServicePath << "as a conflict of" << mServicePath;
	mConflictingServices.append(service);

	// mind it, do not recurse!
	if (!service->mConflictingServices.contains(this))
		service->addConflictingService(this);
}

void Service::beforeDaemonStarted()
{
	foreach (Service *conflictingService, mConflictingServices) {
		// yes this is a bit weird, the setValue triggers a disableDaemon by itself,
		// but we want to make sure the conflicting service is disabled _before_ starting
		// the other service, so it is called here explicitly.
		QLOG_TRACE() << "[Service] stopping conflicting service" << conflictingService->mServicePath;
		conflictingService->disableDaemon();
		// note: only disable the first setting. In theory we can have a conflicting service listening
		// to multiple settings, but we don't have such a service, so changing the first is good enough.
		if (conflictingService->mDbusItemList.count() > 0)
			conflictingService->mDbusItemList.at(0)->setValue(0);
	}
}

void Service::dbusItemChanged(VeQItem *item, QVariant var)
{
	Q_UNUSED(item);
	Q_UNUSED(var);

	QLOG_INFO() << "[Service] DBus service" << mServicePath << "changed:" << item->id() << "-" << var.toString();

	// If one of the services is enabled, start the service.
	QList<VeQItem*>::iterator i;
	bool enable = false;
	for (i = mDbusItemList.begin(); i != mDbusItemList.end(); ++i) {
		QVariant value = (*i)->getValue();

		if (value.isValid()) {
			if (value.toInt() == mStartValue)
				enable = true;
		} else {
			return;
		}
	}

	if (enable)
		enableDaemon();
	else
		disableDaemon();
}

void Service::touch(QString fileName)
{
	QFile file(fileName);
	file.open(QIODevice::ReadWrite);
	file.close();
}

void Service::execSvc(QStringList arguments)
{
	QProcess proc;
	proc.start("svc", arguments);
	proc.waitForFinished();
}
