#include <QFile>
#include "updater.h"
#include "QsLog.h"
#include <QRegExp>

#define UPDATE_FILE "/var/run/swupdate-status"
#define UPDATE_SCRIPT "/opt/victronenergy/swupdate-scripts/check-updates.sh"
#define VERSIONS_FILE "/var/run/versions"
#define SETVERSION_SCRIPT "/opt/victronenergy/swupdate-scripts/set-version.sh"

Updater::Updater(NotificationCenter *notificationCenter, QObject *parent) :
	QObject(parent),
	mRunning(false),
	mUpdateTriggered(false),
	mStatus(0),
	mStatusDescription(tr("Idle")),
	mNotificationCenter(notificationCenter)
{
	// First create file and/or empty it so we know it exists and then add watcher.
	// When user presses check-updates, the update script is called. After checking for updates, the same
	// script will start the upgrade. Before starting the upgrade, it will write the update status and
	// new version.
	touchFile(UPDATE_FILE);
	touchFile(VERSIONS_FILE);

	mUpdateWatcher.addPath(UPDATE_FILE);
	mUpdateWatcher.addPath(VERSIONS_FILE);
	fileChanged(VERSIONS_FILE);
	connect(&mUpdateWatcher, SIGNAL(fileChanged(QString)), this, SLOT(fileChanged(QString)));
	connect(&mUpdateScript, SIGNAL(finished(int, QProcess::ExitStatus)), this, SLOT(updateScriptFinished(int, QProcess::ExitStatus)));
	connect(&mSetVersionScript, SIGNAL(finished(int, QProcess::ExitStatus)), this, SLOT(updateScriptFinished(int, QProcess::ExitStatus)));
}

int Updater::getStatus()
{
	return mStatus;
}

QString Updater::getStatusDescription()
{
	return mStatusDescription;
}

QString Updater::getNotification()
{
	return mNotification;
}

QString Updater::getAvailableUpdate()
{
	return mAvailableUpdate;
}

QString Updater::getOfflineAvailableUpdate()
{
	return mOfflineAvailableUpdate;
}

QString Updater::getAvailableUpdateBuild()
{
	return mAvailableUpdateBuild;
}

QString Updater::getOfflineAvailableUpdateBuild()
{
	return mOfflineAvailableUpdateBuild;
}

QVariantList Updater::getAvailableRootfsVersions()
{
	return mAvailableRootfsVersions;
}

QVariantList Updater::getAvailableRootfsBuildNumbers()
{
	return mAvailableRootfsBuildNumbers;
}

bool Updater::getRunning()
{
	return mRunning;
}

bool Updater::getRunningOffline()
{
	return mRunningOffline;
}

bool Updater::getUdateTriggered()
{
	return mUpdateTriggered;
}

void Updater::getUpdateInfoFromFile(const QString &str)
{
	QFile file(str);

	if (!file.open(QIODevice::ReadOnly | QIODevice::Text))
		return;

	// Line 0 = status, line 1 = "timestamp<space>version"
	mStatus = file.readLine().trimmed().toInt();
	QMap<QString, QString> onlineUpdate = getVersionInfoFromLine(file.readLine());
	QMap<QString, QString> offlineUpdate = getVersionInfoFromLine(file.readLine());

	mAvailableUpdate = onlineUpdate["version"];
	mAvailableUpdateBuild = onlineUpdate["build"];
	mOfflineAvailableUpdate = offlineUpdate["version"];\
	mOfflineAvailableUpdateBuild = offlineUpdate["build"];

	file.close();
}

void Updater::getRootfsInfoFromFile(const QString &str)
{
	QFile file(str);

	QVariantList versions = QVariantList() << "" << "";
	QVariantList buildNumbers = QVariantList() << "" << "";
	QString line;

	if (file.open(QIODevice::ReadOnly | QIODevice::Text)) {
		// Line 0 is always the running version
		// both lines in "timestamp<space>version" format
		line = file.readLine();
		versions[0] = getVersionInfoFromLine(line)["version"];
		buildNumbers[0] =  getVersionInfoFromLine(line)["build"];

		line = file.readLine();
		versions[1] = getVersionInfoFromLine(line)["version"];
		buildNumbers[1] =  getVersionInfoFromLine(line)["build"];

		file.close();
	}

	mAvailableRootfsVersions = versions;
	mAvailableRootfsBuildNumbers = buildNumbers;
	emit availableRootfsChanged();
}

void Updater::fileChanged(const QString &str)
{
	if (str == UPDATE_FILE)
		updateAvailableChanged(str);
	if (str == VERSIONS_FILE)
		getRootfsInfoFromFile(str);
}

void Updater::updateAvailableChanged(const QString& str)
{
	getUpdateInfoFromFile(str);

	mNotification = "";

	QString updatingToVersion = mAvailableUpdate.isEmpty() ? mOfflineAvailableUpdate : mAvailableUpdate;

	QLOG_TRACE() << "ColorControl::updateAvailableChanged: status" << mStatus << "- version"
				 << mAvailableUpdate << "offline version" << mOfflineAvailableUpdate;

	switch (mStatus)
	{
		case Idle:
			mStatusDescription = tr("Idle");
			break;
		case Checking:
			mStatusDescription = tr("Checking...");
			break;
		case DownloadingAndInstalling:
			mStatusDescription = tr("Updating to %1...").arg(updatingToVersion);
			mUpdateTriggered = true;
			if (!mRunning && !mRunningOffline)
				mNotification = tr("Downloading and installing firmware %1").arg(updatingToVersion);
			break;
		case ErrorDuringChecking:
			mStatusDescription = tr("Error during checking");
			break;
		case ErrorDuringUpdating:
			mStatusDescription = tr("Error during updating");
			break;
		case UpdateFileNotFound:
			mStatusDescription = tr("No update file found");
			break;
		case Rebooting:
			mNotification = tr("Firmware update installed, rebooting.");
			break;
		default:
			return;
	}

	emit statusChanged();
	emit runningChanged();
	emit notificationChanged();
	emit availableUpdateChanged();
}

void Updater::checkUpdates(bool offline)
{
	QStringList arguments = QStringList() << "-check";

	if (offline) {
		arguments << "-offline";
		mRunningOffline = true;
	}

	QLOG_TRACE() << "Updater::checkUpdates: ";

	if (mRunning) {
		QLOG_TRACE() << "Updater::checkUpdates: Already running, exit ";
		return;
	}

	mRunning = !offline;
	emit runningChanged();

	QLOG_INFO() << "Updater::checkupdate: starting update script";
	mUpdateScript.start(UPDATE_SCRIPT, arguments);
	if (!mUpdateScript.waitForStarted()) {
		QLOG_ERROR() << "Updater::checkupdate: cannot start update script " << UPDATE_SCRIPT << arguments;
		emit statusChanged();
		mRunning = false;
		mRunningOffline = false;
		emit runningChanged();
	}
}

void Updater::doOfflineUpdate()
{
	doUpdate(true);
}

void Updater::checkOfflineUpdates() {
	checkUpdates(true);
}

void Updater::doUpdate(bool offline)
{
	QStringList arguments = QStringList() << "-update";

	if (offline) {
		arguments << "-offline";
		mRunningOffline = true;
	}

	QLOG_TRACE() << "Updater::doUpdate: ";

	if (mRunning) {
		QLOG_TRACE() << "Updater::doUpdate: Already running, exit ";
		return;
	}

	mRunning = !offline;
	mUpdateTriggered = true;
	emit runningChanged();

	QLOG_INFO() << "Updater::doUpdate: starting update script";
	mUpdateScript.start(UPDATE_SCRIPT, arguments);
	if (!mUpdateScript.waitForStarted()) {
		QLOG_ERROR() << "Updater::doUpdate: cannot start update script " << UPDATE_SCRIPT;
		mRunning = false;
		mRunningOffline = false;
		emit runningChanged();
	}
	return;
}

void Updater::switchVersion()
{
	QStringList arguments = QStringList() << "2";

	if (mRunning) {
		QLOG_TRACE() << "Updater::switchVersion: Already running, exit ";
		return;
	}

	mSetVersionScript.start(SETVERSION_SCRIPT, arguments);

	mRunning = true;
	emit runningChanged();

	if (!mSetVersionScript.waitForStarted()) {
		QLOG_ERROR() << "Updater::setVersion: cannot start setVersion script " << SETVERSION_SCRIPT;
		mRunning = false;
		emit runningChanged();
	}
}

void Updater::updateScriptFinished(int exitCode, QProcess::ExitStatus exitStatus)
{
	Q_UNUSED(exitCode);
	Q_UNUSED(exitStatus);

	QLOG_TRACE() << "Updater::updateScriptFinished: ";

	resetRunningStatus();
}

void Updater::resetRunningStatus()
{
	QLOG_TRACE() << "Updater::resetRunningStatus: ";
	mRunning = false;
	mRunningOffline = false;
	mUpdateTriggered = false;
	emit runningChanged();
}

void Updater::touchFile(QString filePath)
{
	QFile file(filePath);
	if (!file.exists()) {
		file.open(QIODevice::ReadWrite);
		file.resize(0);
		file.close();
	}
}

QMap<QString, QString> Updater::getVersionInfoFromLine(QString versionLine)
{
	QRegExp regexp("(\\d+\\s[v]\\d+\\.\\d+)");
	QMap<QString, QString> version;
	version.insert("build","");
	version.insert("version","");
	if (versionLine.contains(regexp)) {
		version["build"] = versionLine.split(" ")[0].simplified();
		version["version"] = versionLine.split(" ")[1].simplified();
	}
	return version;
}
